# coding: utf-8
"""
jinja2schema.visitors.util
~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import jinja2.nodes

from ..mergers import merge
from ..model import Dictionary, Scalar, Unknown


def visit(node, macroses, config, predicted_struct_cls=Scalar, return_struct_cls=Unknown):
    if isinstance(node, jinja2.nodes.Stmt):
        structure = visit_stmt(node, macroses, config)
    elif isinstance(node, jinja2.nodes.Expr):
        ctx = Context(predicted_struct=predicted_struct_cls.from_ast(node), return_struct_cls=return_struct_cls)
        _, structure = visit_expr(node, ctx, macroses, config)
    elif isinstance(node, jinja2.nodes.Template):
        structure = visit_many(node.body, macroses, config)
    return structure


def visit_many(nodes, macroses, config, predicted_struct_cls=Scalar, return_struct_cls=Unknown):
    """Visits ``nodes`` and merges results.

    :param nodes: list of :class:`jinja2.nodes.Node`
    :param predicted_struct_cls: ``predicted_struct`` for expression visitors will be constructed
                                   using this class by calling :meth:`from_ast` method
    :return: :class:`Dictionary`
    """
    rv = Dictionary()
    for node in nodes:
        if isinstance(node, jinja2.nodes.Extends):
            structure = visit_extends(node, macroses, config, [x for x in nodes if isinstance(x, jinja2.nodes.Block)])
        else:
            structure = visit(node, macroses, config, predicted_struct_cls, return_struct_cls)
        rv = merge(rv, structure)
    return rv


# keep these at the end of file to avoid circular imports
from .expr import Context, visit_expr
from .stmt import visit_stmt, visit_extends



# coding: utf-8
"""
jinja2schema.util
~~~~~~~~~~~~~~~~~
"""
from .model import Dictionary, Scalar, List, Unknown, Tuple


def _format_attrs(var):
    return (u'label={0.label}, required={0.required}, '
            u'constant={0.constant}, linenos={0.linenos}, may_be_d={0.may_be_defined}, '
            u'c_as_u={0.checked_as_undefined}, c_as_d={0.checked_as_defined}').format(var).encode('utf-8')


def _indent(lines, spaces):
    indent = ' ' * spaces
    return [indent + line for line in lines]


def _debug_repr(var):
    rv = []
    if isinstance(var, (Dictionary, Tuple, List)):
        if isinstance(var, Dictionary):
            rv.append('Dictionary({}, {{'.format(_format_attrs(var)))
            content = []
            for key, value in sorted(var.iteritems()):
                key_repr = key + ': '
                value_repr = _debug_repr(value)
                content.append(key_repr + value_repr[0])
                content.extend(_indent(value_repr[1:], spaces=len(key_repr)))
            rv.extend(_indent(content, spaces=4))
            rv.append('})')
        elif isinstance(var, Tuple):
            rv.append('Tuple({},'.format(_format_attrs(var)))
            for el in var.items:
                el_repr = _debug_repr(el)
                el_repr[-1] += ','
                rv.extend(_indent(el_repr, spaces=4))
            rv[-1] = rv[-1][:-1]  # remove comma from the last tuple item
            rv.append(')')
        elif isinstance(var, List):
            rv.append('List({},'.format(_format_attrs(var)))
            rv.extend(_indent(_debug_repr(var.item), spaces=4))
            rv.append(')')
    elif isinstance(var, Scalar):
        rv = ['{}({})'.format(var.__class__.__name__, _format_attrs(var))]
    elif isinstance(var, Unknown):
        rv = ['Unknown({})'.format(_format_attrs(var))]
    return rv


def debug_repr(var):
    """
    :param var: instance of :class:`Variable`
    :return: a string with detailed representation of ``var``
    """
    return '\n'.join(_debug_repr(var))


# coding: utf-8
"""
jinja2schema.visitors.expr
~~~~~~~~~~~~~~~~~~~~~~~~~~

Expression is an instance of :class:`jinja2.nodes.Expr`.
Expression visitors return a tuple which contains expression type and expression structure.
"""
import functools

from jinja2 import nodes

from ..model import Scalar, Dictionary, List, Unknown, Tuple, String, Number, Boolean
from ..mergers import merge_rtypes, merge, merge_many, merge_bool_expr_structs
from ..exceptions import InvalidExpression, UnexpectedExpression, MergeException
from ..config import default_config
from .. import _compat
from .util import visit_many


class Context(object):
    """
    Context is used when parsing expressions.

    Suppose there is an expression::

        {{ data.field.subfield }}

    It has the following AST::

        Getattr(
            node=Getattr(
                node=Name(name='data')
                attr='field'
            ),
            attr='subfield'
        )

    :func:`visit_getattr` returns a pair that looks like this::

        (
            # return type:
            Scalar(...),
            # structure:
            {
                'data: {
                    'field': {
                        'subfield': Scalar(...)
                    }
                }
            }
        }

    The return type is defined by the outermost :class:`nodes.Getattr` node, which
    in this case is being printed.
    The structure is build during AST traversal from outer to inners nodes and it is
    kind of "reversed" in relation to the AST.
    :class:`Context` is intended for:

    * capturing a return type and passing it to the innermost expression node;
    * passing a structure "under construction" to the visitors of nested nodes.

    Let's look through an example.

    Suppose :func:`visit_getattr` is called with the following arguments::

       ast = Getattr(node=Getattr(node=Name(name='data'), attr='field'), attr='subfield'))
       context = Context(return_struct_cls=Scalar, predicted_struct=Scalar())

    It looks to the outermost AST node and based on it's type (which is :class:`nodes.Getattr`)
    and it's ``attr`` field (which equals to ``"subfield"``) infers that a variable described by the
    nested AST node must a dictionary with ``"subfield"`` key.

    It calls a visitor for inner node and :func:`visit_getattr` gets called again, but
    with different arguments::

       ast = Getattr(node=Name(name='data', ctx='load'), attr='field')
       ctx = Context(return_struct_cls=Scalar, predicted_struct=Dictionary({subfield: Scalar()}))

    :func:`visit_getattr` applies the same logic again. The inner node is a :class:`nodes.Name`, so that
    it calls :func:`visit_name` with the following arguments::

       ast = Name(name='data')
       ctx = Context(
           return_struct_cls=Scalar,
           predicted_struct=Dictionary({
               field: Dictionary({subfield: Scalar()}))
           })
       )

    :func:`visit_name` does not do much by itself. Based on a context it knows what structure and
    what type must have a variable described by a given :class:`nodes.Name` node, so
    it just returns a pair::

        (instance of context.return_struct_cls, Dictionary({data: context.predicted_struct}})
    """
    def __init__(self, ctx=None, return_struct_cls=None, predicted_struct=None):
        self.predicted_struct = None
        self.return_struct_cls = Unknown
        if ctx:
            self.predicted_struct = ctx.predicted_struct
            self.return_struct_cls = ctx.return_struct_cls
        if predicted_struct:
            self.predicted_struct = predicted_struct
        if return_struct_cls:
            self.return_struct_cls = return_struct_cls

    def get_predicted_struct(self, label=None):
        rv = self.predicted_struct.clone()
        if label:
            rv.label = label
        return rv

    def meet(self, actual_struct, actual_ast):
        try:
            merge(self.predicted_struct, actual_struct)
        except MergeException:
            raise UnexpectedExpression(self.predicted_struct, actual_ast, actual_struct)
        else:
            return True


expr_visitors = {}


def visits_expr(node_cls):
    """Decorator that registers a function as a visitor for ``node_cls``.

    :param node_cls: subclass of :class:`jinja2.nodes.Expr`
    """
    def decorator(func):
        expr_visitors[node_cls] = func
        @functools.wraps(func)
        def wrapped_func(ast, ctx, macroses=None, config=default_config):
            assert isinstance(ast, node_cls)
            return func(ast, ctx, macroses=macroses, config=config)
        return wrapped_func
    return decorator


def visit_expr(ast, ctx, macroses=None, config=default_config):
    """Returns a structure of ``ast``.

    :param ctx: :class:`Context`
    :param ast: instance of :class:`jinja2.nodes.Expr`
    :returns: a tuple where the first element is an expression type (instance of :class:`Variable`)
              and the second element is an expression structure (instance of :class:`.model.Dictionary`)
    """
    visitor = expr_visitors.get(type(ast))
    if not visitor:
        for node_cls, visitor_ in _compat.iteritems(expr_visitors):
            if isinstance(ast, node_cls):
                visitor = visitor_
    if not visitor:
        raise Exception('expression visitor for {0} is not found'.format(type(ast)))
    return visitor(ast, ctx, macroses, config=config)


def _visit_dict(ast, ctx, macroses, items, config=default_config):
    """A common logic behind nodes.Dict and nodes.Call (``{{ dict(a=1) }}``)
    visitors.

    :param items: a list of (key, value); key may be either AST node or string
    """
    ctx.meet(Dictionary(), ast)
    rtype = Dictionary.from_ast(ast, constant=True)
    struct = Dictionary()
    for key, value in items:
        value_rtype, value_struct = visit_expr(value, Context(
            predicted_struct=Unknown.from_ast(value)), macroses, config=config)
        struct = merge(struct, value_struct)
        if isinstance(key, nodes.Node):
            key_rtype, key_struct = visit_expr(key, Context(predicted_struct=Scalar.from_ast(key)), macroses, config=config)
            struct = merge(struct, key_struct)
            if isinstance(key, nodes.Const):
                rtype[key.value] = value_rtype
        elif isinstance(key, _compat.string_types):
            rtype[key] = value_rtype
    return rtype, struct


@visits_expr(nodes.BinExpr)
def visit_bin_expr(ast, ctx, macroses=None, config=default_config):
    l_rtype, l_struct = visit_expr(ast.left, ctx, macroses, config=config)
    r_rtype, r_struct = visit_expr(ast.right, ctx, macroses, config=config)
    rv = merge_bool_expr_structs(l_struct, r_struct)
    return merge_rtypes(l_rtype, r_rtype, operator=ast.operator), rv


@visits_expr(nodes.UnaryExpr)
def visit_unary_expr(ast, ctx, macroses=None, config=default_config):
    return visit_expr(ast.node, ctx, macroses, config=config)


@visits_expr(nodes.Compare)
def visit_compare(ast, ctx, macroses=None, config=default_config):
    ctx.meet(Boolean(), ast)
    rtype, struct = visit_expr(ast.expr, Context(predicted_struct=Unknown.from_ast(ast.expr)),
                               macroses, config=config)
    for op in ast.ops:
        op_rtype, op_struct = visit_expr(op.expr, Context(predicted_struct=Unknown.from_ast(ast.expr)),
                                         macroses, config=config)
        struct = merge(struct, op_struct)
    return Boolean.from_ast(ast), struct


@visits_expr(nodes.Slice)
def visit_slice(ast, ctx, macroses=None, config=default_config):
    nodes = [node for node in [ast.start, ast.stop, ast.step] if node is not None]
    struct = visit_many(nodes, macroses, config,
                        predicted_struct_cls=Number,
                        return_struct_cls=Number)
    return Unknown(), struct


@visits_expr(nodes.Name)
def visit_name(ast, ctx, macroses=None, config=default_config):
    kwargs = {}
    return ctx.return_struct_cls.from_ast(ast, **kwargs), Dictionary({
        ast.name: ctx.get_predicted_struct(label=ast.name)
    })


@visits_expr(nodes.Getattr)
def visit_getattr(ast, ctx, macroses=None, config=default_config):
    context = Context(
        ctx=ctx,
        predicted_struct=Dictionary.from_ast(ast, {
            ast.attr: ctx.get_predicted_struct(label=ast.attr),
        }))
    return visit_expr(ast.node, context, macroses, config=config)


@visits_expr(nodes.Getitem)
def visit_getitem(ast, ctx, macroses=None, config=default_config):
    arg = ast.arg
    if isinstance(arg, nodes.Const):
        if isinstance(arg.value, int):
            if config.TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE == 'list':
                predicted_struct = List.from_ast(ast, ctx.get_predicted_struct())
            elif config.TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE == 'dictionary':
                predicted_struct = Dictionary.from_ast(ast, {
                    arg.value: ctx.get_predicted_struct(),
                })
            elif config.TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE == 'tuple':
                items = [Unknown() for i in range(arg.value + 1)]
                items[arg.value] = ctx.get_predicted_struct()
                predicted_struct = Tuple.from_ast(ast, tuple(items), may_be_extended=True)
        elif isinstance(arg.value, _compat.string_types):
            predicted_struct = Dictionary.from_ast(ast, {
                arg.value: ctx.get_predicted_struct(label=arg.value),
            })
        else:
            raise InvalidExpression(arg, '{0} is not supported as an index for a list or'
                                         ' a key for a dictionary'.format(arg.value))
    elif isinstance(arg, nodes.Slice):
        predicted_struct = List.from_ast(ast, ctx.get_predicted_struct())
    else:
        if config.TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE == 'list':
            predicted_struct = List.from_ast(ast, ctx.get_predicted_struct())
        elif config.TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE == 'dictionary':
            predicted_struct = Dictionary.from_ast(ast)

    _, arg_struct = visit_expr(arg, Context(predicted_struct=Scalar.from_ast(arg)), macroses, config=config)
    rtype, struct = visit_expr(ast.node, Context(
        ctx=ctx,
        predicted_struct=predicted_struct), macroses, config=config)
    return rtype, merge(struct, arg_struct)


@visits_expr(nodes.Test)
def visit_test(ast, ctx, macroses=None, config=default_config):
    ctx.meet(Boolean(), ast)
    if ast.name in ('divisibleby', 'escaped', 'even', 'lower', 'odd', 'upper'):
        # TODO
        predicted_struct = Scalar.from_ast(ast.node)
    elif ast.name in ('defined', 'undefined', 'equalto', 'iterable', 'mapping',
                      'none', 'number', 'sameas', 'sequence', 'string'):
        predicted_struct = Unknown.from_ast(ast.node)
        if ast.name == 'defined':
            predicted_struct.checked_as_defined = True
        elif ast.name == 'undefined':
            predicted_struct.checked_as_undefined = True
    else:
        raise InvalidExpression(ast, 'unknown test "{0}"'.format(ast.name))
    rtype, struct = visit_expr(ast.node, Context(return_struct_cls=Boolean,
                                                 predicted_struct=predicted_struct), macroses, config=config)
    if ast.name == 'divisibleby':
        if not ast.args:
            raise InvalidExpression(ast, 'divisibleby must have an argument')
        _, arg_struct = visit_expr(ast.args[0],
                                   Context(predicted_struct=Number.from_ast(ast.args[0])), macroses, config=config)
        struct = merge(arg_struct, struct)
    return rtype, struct


@visits_expr(nodes.Concat)
def visit_concat(ast, ctx, macroses=None, config=default_config):
    ctx.meet(Scalar(), ast)
    return String.from_ast(ast), visit_many(ast.nodes, macroses, config, predicted_struct_cls=String)


@visits_expr(nodes.CondExpr)
def visit_cond_expr(ast, ctx, macroses=None, config=default_config):
    if config.BOOLEAN_CONDITIONS:
        test_predicted_struct = Boolean.from_ast(ast.test)
    else:
        test_predicted_struct = Unknown.from_ast(ast.test)
    test_rtype, test_struct = visit_expr(ast.test, Context(predicted_struct=test_predicted_struct), macroses, config=config)
    if_rtype, if_struct = visit_expr(ast.expr1, ctx, macroses, config=config)
    else_rtype, else_struct = visit_expr(ast.expr2, ctx, macroses, config=config)
    struct = merge_many(if_struct, test_struct, else_struct)
    rtype = merge_rtypes(if_rtype, else_rtype)

    for var_name, var_struct in test_struct.iteritems():
        if var_struct.checked_as_defined or var_struct.checked_as_undefined:
            if var_struct.checked_as_undefined:
                lookup_struct = if_struct
            elif var_struct.checked_as_defined:
                lookup_struct = else_struct
            struct[var_name].may_be_defined = (lookup_struct and
                                               var_name in lookup_struct and
                                               lookup_struct[var_name].constant)
            struct[var_name].checked_as_defined = test_struct[var_name].checked_as_defined and (
                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
            )
            struct[var_name].checked_as_undefined = test_struct[var_name].checked_as_undefined and (
                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
            )

    return rtype, struct


@visits_expr(nodes.Call)
def visit_call(ast, ctx, macroses=None, config=default_config):
    if isinstance(ast.node, nodes.Name):
        if macroses and ast.node.name in macroses:
            macro = macroses[ast.node.name]
            call = MacroCall(macro, ast.args, ast.kwargs, config=config)
            args_struct = call.match_passed_args_to_expected_args()
            if call.passed_args:
                args_struct = merge(args_struct, call.match_passed_args_to_expected_kwargs())
            if call.passed_kwargs:
                args_struct = merge(args_struct, call.match_passed_kwargs_to_expected_args())
            if call.passed_kwargs:
                args_struct = merge(args_struct, call.match_passed_kwargs_to_expected_kwargs())

            if call.passed_args or call.expected_args:
                raise InvalidExpression(ast, ('incorrect usage of "{0}". it takes '
                                              'exactly {1} positional arguments'.format(macro.name, len(macro.args))))
            if call.passed_kwargs:
                first_unknown_kwarg = next(_compat.iterkeys(call.passed_kwargs))
                raise InvalidExpression(ast, ('incorrect usage of "{0}". unknown keyword argument '
                                              '"{1}" is passed'.format(macro.name, first_unknown_kwarg)))
            return Unknown(), args_struct
        elif ast.node.name == 'range':
            ctx.meet(List(Unknown()), ast)
            struct = Dictionary()
            for arg in ast.args:
                arg_rtype, arg_struct = visit_expr(arg, Context(
                    predicted_struct=Number.from_ast(arg)), macroses, config=config)
                struct = merge(struct, arg_struct)
            return List(Number()), struct
        elif ast.node.name == 'lipsum':
            ctx.meet(Scalar(), ast)
            struct = Dictionary()
            # probable TODO: set possible types for args and kwargs
            for arg in ast.args:
                arg_rtype, arg_struct = visit_expr(arg, Context(predicted_struct=Scalar.from_ast(arg)), macroses, config=config)
                struct = merge(struct, arg_struct)
            for kwarg in ast.kwargs:
                arg_rtype, arg_struct = visit_expr(kwarg.value, Context(predicted_struct=Scalar.from_ast(kwarg)), macroses, config=config)
                struct = merge(struct, arg_struct)
            return String(), struct
        elif ast.node.name == 'dict':
            ctx.meet(Dictionary(), ast)
            if ast.args:
                raise InvalidExpression(ast, 'dict accepts only keyword arguments')
            return _visit_dict(ast, ctx, macroses, [(kwarg.key, kwarg.value) for kwarg in ast.kwargs], config=config)
        else:
            raise InvalidExpression(ast, '"{0}" call is not supported'.format(ast.node.name))
    elif isinstance(ast.node, nodes.Getattr):
        if ast.node.attr in ('keys', 'iterkeys', 'values', 'itervalues'):
            ctx.meet(List(Unknown()), ast)
            rtype, struct = visit_expr(
                ast.node.node, Context(predicted_struct=Dictionary.from_ast(ast.node.node)), macroses, config=config)
            return List(Unknown()), struct
        if ast.node.attr in ('startswith', 'endswith'):
            ctx.meet(Boolean(), ast)
            rtype, struct = visit_expr(
                ast.node.node, Context(predicted_struct=String.from_ast(ast.node.node)), macroses, config=config)
            return Boolean(), struct
        if ast.node.attr == 'split':
            ctx.meet(List(String()), ast)
            rtype, struct = visit_expr(
                ast.node.node, Context(predicted_struct=String.from_ast(ast.node.node)), macroses, config=config)
            if ast.args:
                arg = ast.args[0]
                _, arg_struct = visit_expr(arg, Context(predicted_struct=String.from_ast(arg)), macroses, config=config)
                struct = merge(struct, arg_struct)
            return List(String()), struct
        raise InvalidExpression(ast, '"{0}" call is not supported'.format(ast.node.attr))


@visits_expr(nodes.Filter)
def visit_filter(ast, ctx, macroses=None, config=default_config):
    return_struct_cls = None
    if ast.name in ('abs', 'striptags', 'capitalize', 'center', 'escape', 'filesizeformat',
                    'float', 'forceescape', 'format', 'indent', 'int', 'replace', 'round',
                    'safe', 'string', 'striptags', 'title', 'trim', 'truncate', 'upper',
                    'urlencode', 'urlize', 'wordcount', 'wordwrap', 'e'):
        ctx.meet(Scalar(), ast)
        if ast.name in ('abs', 'round'):
            node_struct = Number.from_ast(ast.node)
            return_struct_cls = Number
        elif ast.name in ('float', 'int'):
            node_struct = Scalar.from_ast(ast.node)
            return_struct_cls = Number
        elif ast.name in ('striptags', 'capitalize', 'center', 'escape', 'forceescape', 'format', 'indent',
                          'replace', 'safe', 'title', 'trim', 'truncate', 'upper', 'urlencode',
                          'urlize', 'wordwrap', 'e'):
            node_struct = String.from_ast(ast.node)
            return_struct_cls = String
        elif ast.name == 'filesizeformat':
            node_struct = Number.from_ast(ast.node)
            return_struct_cls = String
        elif ast.name == 'string':
            node_struct = Scalar.from_ast(ast.node)
            return_struct_cls = String
        elif ast.name == 'wordcount':
            node_struct = String.from_ast(ast.node)
            return_struct_cls = Number
        else:
            node_struct = Scalar.from_ast(ast.node)
    elif ast.name in ('batch', 'slice'):
        ctx.meet(List(List(Unknown())), ast)
        rtype = List(List(Unknown(), linenos=[ast.node.lineno]), linenos=[ast.node.lineno])
        node_struct = merge(rtype, ctx.get_predicted_struct()).item
        _, struct = visit_expr(ast.node, Context(
            ctx=ctx,
            return_struct_cls=return_struct_cls,
            predicted_struct=node_struct
        ), macroses, config=config)
        return rtype, struct
    elif ast.name == 'default':
        default_value_rtype, default_value_struct = visit_expr(
            ast.args[0], Context(predicted_struct=Unknown.from_ast(ast.args[0])), macroses, config=config)
        node_struct = merge(
            ctx.get_predicted_struct(),
            default_value_rtype,
        )
        node_struct.used_with_default = True
        node_struct.value = default_value_rtype.value
    elif ast.name == 'dictsort':
        ctx.meet(List(Tuple([Scalar(), Unknown()])), ast)
        node_struct = Dictionary.from_ast(ast.node)
    elif ast.name == 'join':
        ctx.meet(Scalar(), ast)
        node_struct = List.from_ast(ast.node, String())
        rtype, struct = visit_expr(ast.node, Context(
            return_struct_cls=String,
            predicted_struct=node_struct
        ), macroses, config=config)
        arg_rtype, arg_struct = visit_expr(ast.args[0],
                                           Context(predicted_struct=String.from_ast(ast.args[0])), macroses, config=config)
        return rtype, merge(struct, arg_struct)
    elif ast.name in ('first', 'last', 'random', 'length', 'sum'):
        if ast.name in ('first', 'last', 'random'):
            el_struct = ctx.get_predicted_struct()
        elif ast.name == 'length':
            ctx.meet(Scalar(), ast)
            return_struct_cls = Number
            el_struct = Unknown()
        else:
            ctx.meet(Scalar(), ast)
            el_struct = Scalar()
        node_struct = List.from_ast(ast.node, el_struct)
    elif ast.name in ('groupby', 'map', 'reject', 'rejectattr', 'select', 'selectattr', 'sort'):
        ctx.meet(List(Unknown()), ast)
        node_struct = merge(
            List(Unknown()),
            ctx.get_predicted_struct()
        )
    elif ast.name == 'list':
        ctx.meet(List(Scalar()), ast)
        node_struct = merge(
            List(Scalar.from_ast(ast.node)),
            ctx.get_predicted_struct()
        ).item
    elif ast.name == 'pprint':
        ctx.meet(Scalar(), ast)
        node_struct = ctx.get_predicted_struct()
    elif ast.name == 'xmlattr':
        ctx.meet(Scalar(), ast)
        node_struct = Dictionary.from_ast(ast.node)
    elif ast.name == 'attr':
        raise InvalidExpression(ast, 'attr filter is not supported')
    else:
        raise InvalidExpression(ast, 'unknown filter')
    rv = visit_expr(ast.node, Context(
        ctx=ctx,
        return_struct_cls=return_struct_cls,
        predicted_struct=node_struct
    ), macroses, config=config)
    return rv


# :class:`nodes.Literal` visitors

@visits_expr(nodes.TemplateData)
def visit_template_data(ast, ctx, macroses=None, config=default_config):
    return Scalar(), Dictionary()


@visits_expr(nodes.Const)
def visit_const(ast, ctx, macroses=None, config=default_config):
    ctx.meet(Scalar(), ast)
    if isinstance(ast.value, _compat.string_types):
        rtype = String.from_ast(ast, constant=True)
    elif isinstance(ast.value, bool):
        rtype = Boolean.from_ast(ast, constant=True)
    elif isinstance(ast.value, (int, float, complex)):
        rtype = Number.from_ast(ast, constant=True)
    else:
        rtype = Scalar.from_ast(ast, constant=True)
    return rtype, Dictionary()


@visits_expr(nodes.Tuple)
def visit_tuple(ast, ctx, macroses=None, config=default_config):
    ctx.meet(Tuple(None), ast)

    struct = Dictionary()
    item_structs = []
    for item in ast.items:
        item_rtype, item_struct = visit_expr(item, ctx, macroses, config=config)
        item_structs.append(item_rtype)
        struct = merge(struct, item_struct)
    rtype = Tuple.from_ast(ast, item_structs, constant=True)
    return rtype, struct


@visits_expr(nodes.List)
def visit_list(ast, ctx, macroses=None, config=default_config):
    ctx.meet(List(Unknown()), ast)
    struct = Dictionary()

    predicted_struct = merge(List(Unknown()), ctx.get_predicted_struct()).item
    el_rtype = None
    for item in ast.items:
        item_rtype, item_struct = visit_expr(item, Context(predicted_struct=predicted_struct), macroses, config=config)
        struct = merge(struct, item_struct)
        if el_rtype is None:
            el_rtype = item_rtype
        else:
            el_rtype = merge_rtypes(el_rtype, item_rtype)
    rtype = List.from_ast(ast, el_rtype or Unknown(), constant=True)
    return rtype, struct


@visits_expr(nodes.Dict)
def visit_dict(ast, ctx, macroses=None, config=default_config):
    ctx.meet(Dictionary(), ast)
    return _visit_dict(ast, ctx, macroses, [(item.key, item.value) for item in ast.items], config=config)


from ..macro import MacroCall



# coding: utf-8
"""
jinja2schema.macro
~~~~~~~~~~~~~~~~~~
"""
from . import _compat
from .config import default_config
from .mergers import merge
from .model import Unknown, Dictionary


class Macro(object):
    """A macro.

    .. attribute:: name

        Name.

    .. attribute:: args

        Positional arguments. A list of 2-tuples whose first element is string that
        contains argument name and second is a :class:`Variable` -- a structure which
        expected of that argument.
        Arguments must be in the same order as they are listed in macro signature.

    .. attribute:: kwargs

        The same as :attr:`args`, but keyword arguments.
    """
    def __init__(self, name, args, kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs


class MacroCall(object):
    def __init__(self, macro, passed_args, passed_kwargs, config=default_config):
        self.config = config

        self.passed_args = []
        for arg_ast in passed_args:
            arg_rtype, arg_struct = visit_expr(
                arg_ast, Context(predicted_struct=Unknown.from_ast(arg_ast)), config=config)
            self.passed_args.append((arg_ast, arg_rtype))

        self.passed_kwargs = {}
        for kwarg_ast in passed_kwargs:
            kwarg_rtype, kwarg_struct = visit_expr(
                kwarg_ast.value, Context(predicted_struct=Unknown.from_ast(kwarg_ast)), config=config)
            self.passed_kwargs[kwarg_ast.key] = (kwarg_ast, kwarg_rtype)

        self.expected_args = macro.args[:]
        self.expected_kwargs = macro.kwargs[:]

    def _match_passed_args(self, to_args):
        rv = Dictionary()
        matched_args = list(zip(self.passed_args, to_args))
        for i, ((arg_ast, arg), (expected_arg_name, expected_arg)) in enumerate(matched_args, start=1):
            _, s = visit_expr(arg_ast, Context(predicted_struct=merge(arg, expected_arg)), config=self.config)
            rv = merge(rv, s)
        del self.passed_args[:len(matched_args)]
        del to_args[:len(matched_args)]
        return rv

    def match_passed_args_to_expected_args(self):
        return self._match_passed_args(self.expected_args)

    def match_passed_args_to_expected_kwargs(self):
        return self._match_passed_args(self.expected_kwargs)

    def _match_passed_kwargs(self, to_args):
        rv = Dictionary()
        for kwarg_name, (kwarg_ast, kwarg_type) in list(_compat.iteritems(self.passed_kwargs)):
            for (expected_arg_name, expected_arg_struct) in list(to_args):
                if kwarg_name == expected_arg_name:
                    _, s = visit_expr(kwarg_ast.value,
                                      Context(predicted_struct=merge(kwarg_type, expected_arg_struct)),
                                      config=self.config)
                    rv = merge(rv, s)
                    to_args.remove((expected_arg_name, expected_arg_struct))
                    del self.passed_kwargs[kwarg_name]
        return rv

    def match_passed_kwargs_to_expected_args(self):
        return self._match_passed_kwargs(self.expected_args)

    def match_passed_kwargs_to_expected_kwargs(self):
        return self._match_passed_kwargs(self.expected_kwargs)


from .visitors.expr import visit_expr
from .visitors.expr import Context



# coding: utf-8
"""
jinja2schema.mergers
~~~~~~~~~~~~~~~~~~~~
"""
import itertools
from jinja2schema.util import debug_repr

from .model import Scalar, Dictionary, List, Unknown, Tuple
from .exceptions import MergeException
from ._compat import zip_longest


def merge(fst, snd, custom_merger=None):
    """Merges two variables.

    :param fst: first variable
    :type fst: :class:`.model.Variable`
    :param snd: second variable
    :type snd: :class:`.model.Variable`

    .. note::

        ``fst`` must reflect expressions that occur in template **before** the expressions of ``snd``.
    """

    # useful for debugging:
    #
    # assert (not (fst.linenos and snd.linenos) or
    #         (fst.linenos == snd.linenos) or  # TODO this is a hack
    #         max(fst.linenos) <= min(snd.linenos))

    if isinstance(fst, Unknown):
        result = snd.clone()
    elif isinstance(snd, Unknown):
        result = fst.clone()
    elif isinstance(fst, Scalar) and isinstance(snd, Scalar):
        fst_type = type(fst)
        snd_type = type(snd)
        if issubclass(fst_type, snd_type):
            result = fst_type()
        elif issubclass(snd_type, fst_type):
            result = snd_type()
        else:
            raise MergeException(fst, snd)
    elif isinstance(fst, Dictionary) and isinstance(snd, Dictionary):
        result = Dictionary()
        for k in set(itertools.chain(fst.iterkeys(), snd.iterkeys())):
            if k in fst and k in snd:
                result[k] = merge(fst[k], snd[k], custom_merger=custom_merger)
            elif k in fst:
                result[k] = fst[k].clone()
            elif k in snd:
                result[k] = snd[k].clone()
    elif isinstance(fst, List) and isinstance(snd, List):
        result = List(merge(fst.item, snd.item, custom_merger=custom_merger))
    elif isinstance(fst, Tuple) and isinstance(snd, Tuple):
        if fst.items is snd.items is None:
            result = Tuple(None)
        else:
            if len(fst.items) != len(snd.items) and not (fst.may_be_extended or snd.may_be_extended):
                raise MergeException(fst, snd)
            result = Tuple([merge(a, b, custom_merger=custom_merger)
                            for a, b in zip_longest(fst.items, snd.items, fillvalue=Unknown())])
    else:
        raise MergeException(fst, snd)
    result.label = fst.label or snd.label
    result.linenos = list(sorted(set(fst.linenos + snd.linenos)))
    result.constant = fst.constant
    result.may_be_defined = fst.may_be_defined
    result.used_with_default = fst.used_with_default and snd.used_with_default
    result.checked_as_defined = fst.checked_as_defined and snd.checked_as_defined
    result.checked_as_undefined = fst.checked_as_undefined and snd.checked_as_undefined
    if fst.value == snd.value:
        result.value = fst.value
    if callable(custom_merger):
        result = custom_merger(fst, snd, result)
    return result


def merge_many(fst, snd, *args):
    struct = merge(fst, snd)
    if args:
        return merge_many(struct, *args)
    else:
        return struct


def merge_bool_expr_structs(fst, snd, operator=None):
    def merger(fst, snd, result):
        result.checked_as_defined = fst.checked_as_defined
        result.checked_as_undefined = fst.checked_as_undefined and snd.checked_as_undefined
        return result
    return merge(fst, snd, custom_merger=merger)


def merge_rtypes(fst, snd, operator=None):
    if operator in ('+', '-'):
        if type(fst) is not type(snd) and not (isinstance(fst, Unknown) or isinstance(snd, Unknown)):
            raise MergeException(fst, snd)
    return merge(fst, snd)



# coding: utf-8
"""

jinja2schema
============

Type inference for Jinja2 templates.

See https://jinja2schema.readthedocs.io/ for documentation.

:copyright: (c) 2017 Anton Romanovich
:license: BSD

"""

__title__ = 'jinja2schema'
__author__ = 'Anton Romanovich'
__license__ = 'BSD'
__copyright__ = 'Copyright 2017 Anton Romanovich'
__version__ = '0.1.3'
__version_info__ = tuple(int(i) for i in __version__.split('.'))


from .config import Config
from .core import (parse, infer, infer_from_ast, to_json_schema,
                   JSONSchemaDraft4Encoder, StringJSONSchemaDraft4Encoder)
from .exceptions import InferException, MergeException, InvalidExpression, UnexpectedExpression



# coding: utf-8
"""
jinja2schema.core
~~~~~~~~~~~~~~~~~
"""
import jinja2

from .config import Config
from .model import Dictionary, List, Tuple, Scalar, Number, Unknown, Boolean, String
from .visitors import visit
from . import _compat


def parse(template, jinja2_env=None):
    """Parses Jinja2 template and returns it's AST.

    :type template: basestring
    :type jinja2_env: :class:`jinja2.Environment`
    :rtype: :class:`jinja2.nodes.Template`
    """
    if jinja2_env is None:
        jinja2_env = jinja2.Environment()
    return jinja2_env.parse(template)


def _ignore_constants(var):
    if isinstance(var, Dictionary):
        for k, v in list(_compat.iteritems(var)):
            if v.constant and not v.may_be_defined:
                del var[k]
            else:
                _ignore_constants(v)
    return var


def infer_from_ast(ast, ignore_constants=True, config=Config()):
    """Returns a :class:`.model.Dictionary` which reflects a structure of variables used
    within ``ast``.

    :param ast: AST
    :type ast: :class:`jinja2.nodes.Node`
    :param ignore_constants: excludes constant variables from a resulting structure
    :param config: a config
    :type config: :class:`.config.Config`
    :rtype: :class:`.model.Dictionary`
    :raises: :class:`.exceptions.MergeException`, :class:`.exceptions.InvalidExpression`,
             :class:`.exceptions.UnexpectedExpression`
    """
    rv = visit(ast, {}, config)
    if ignore_constants:
        rv = _ignore_constants(rv)
    return rv


def infer(template, config=Config()):
    """Returns a :class:`.model.Dictionary` which reflects a structure of the context required by ``template``.

    :param template: a template
    :type template: string
    :param config: a config
    :type config: :class:`.config.Config`
    :rtype: :class:`.model.Dictionary`
    :raises: :class:`.exceptions.MergeException`, :class:`.exceptions.InvalidExpression`,
             :class:`.exceptions.UnexpectedExpression`
    """
    return infer_from_ast(parse(template), config=config, ignore_constants=True)


class JSONSchemaDraft4Encoder(object):
    """Extensible JSON schema encoder for :class:`.model.Variable`."""

    def encode_common_attrs(self, var):
        """Returns a subset of JSON schema of variable `var` that describes
        attributes that are common for all the variable types, such as label.
        """
        rv = {}
        if var.label:
            rv['title'] = var.label
        if var.value and var.used_with_default:
            rv['default'] = var.value
        return rv

    def encode(self, var):
        """Returns a JSON schema of variable `var`.

        :type var: :class:`.model.Variable`
        :rtype: :class:`dict`
        """
        rv = self.encode_common_attrs(var)
        if isinstance(var, Dictionary):
            rv.update({
                'type': 'object',
                'properties': dict((k, self.encode(v)) for k, v in var.iteritems()),
                'required': [k for k, v in var.iteritems() if v.required],
            })
        elif isinstance(var, List):
            rv.update({
                'type': 'array',
                'items': self.encode(var.item),
            })
        elif isinstance(var, Tuple):
            rv.update({
                'type': 'array',
                'items': [self.encode(item) for item in var.items],
            })
        elif isinstance(var, Unknown):
            rv['anyOf'] = [
                {'type': 'object'},
                {'type': 'array'},
                {'type': 'string'},
                {'type': 'number'},
                {'type': 'boolean'},
                {'type': 'null'},
            ]
        elif isinstance(var, String):
            rv['type'] = 'string'
        elif isinstance(var, Number):
            rv['type'] = 'number'
        elif isinstance(var, Boolean):
            rv['type'] = 'boolean'
        elif isinstance(var, Scalar):
            rv['anyOf'] = [
                {'type': 'boolean'},
                {'type': 'null'},
                {'type': 'number'},
                {'type': 'string'},
            ]
        return rv


class StringJSONSchemaDraft4Encoder(JSONSchemaDraft4Encoder):
    """Encodes :class:`.model.Unknown` and :class:`.model.Scalar` (but not it's subclasses --
    :class:`.model.String`, :class:`.model.Number` or :class:`.model.Boolean`) variables as strings.

    Useful for rendering forms using resulting JSON schema, as most of form-rendering
    tools do not support "anyOf" validator.
    """
    def encode(self, var):
        if isinstance(var, Unknown) or type(var) is Scalar:
            rv = self.encode_common_attrs(var)
            rv['type'] = 'string'
        else:
            rv = super(StringJSONSchemaDraft4Encoder, self).encode(var)
        return rv


def to_json_schema(var, jsonschema_encoder=JSONSchemaDraft4Encoder):
    """Returns JSON schema that describes ``var``.

    :param var: a variable
    :type var: :class:`.model.Variable`
    :param jsonschema_encoder: JSON schema encoder
    :type jsonschema_encoder: subclass of :class:`JSONSchemaEncoder`
    :return: :class:`dict`
    """
    return jsonschema_encoder().encode(var)



# coding: utf-8
"""
jinja2schema.model
~~~~~~~~~~~~~~~~~~
"""
import pprint

from jinja2 import nodes

from . import _compat


class Variable(object):
    """A base variable class.

    .. attribute:: linenos

        An ordered list of line numbers on which the variable occurs.

    .. attribute:: label

        A name of the variable in template.

    .. attribute:: constant

        Is true if the variable is defined using a ``{% set %}`` tag before it
        occurs within any other statement in the template, or inferred from
        a constant Jinja2 node.

    .. attribute:: may_be_defined

        Is true if the variable would be defined (using a ``{% set %}`` expression) if it's missing
        from the template context. For example, ``x`` is ``may_be_defined`` in the following template::

            {% if x is undefined %} {% set x = 1 %} {% endif %}

    .. attribute:: used_with_default

        Is true if the variable occurs only within the ``default`` filter.

    .. attribute:: checked_as_undefined

        Is true if the variable occurs within ``{% if %}`` block which condition checks
        if the variable is undefined.

    .. attribute:: checked_as_defined

        Is true if the variable occurs within ``{% if %}`` block which condition checks
        if the variable is defined.

    .. attribute:: value

        Value of the variable in template. Set by default filter or assignment.
    """
    def __init__(self, label=None, linenos=None, constant=False,
                 may_be_defined=False, used_with_default=False,
                 checked_as_undefined=False, checked_as_defined=False, value=None):
        self.label = label
        self.linenos = linenos if linenos is not None else []
        self.constant = constant
        self.may_be_defined = may_be_defined
        self.used_with_default = used_with_default
        self.checked_as_undefined = checked_as_undefined
        self.checked_as_defined = checked_as_defined
        self.value = value

    def clone(self):
        cls = type(self)
        return cls(**self.__dict__)

    @classmethod
    def _get_kwargs_from_ast(cls, ast):
        return {
            'linenos': [ast.lineno],
            'label': ast.name if isinstance(ast, nodes.Name) else None,
            'value': ast.value if hasattr(ast, 'value') else None,
        }

    @classmethod
    def from_ast(cls, ast, **kwargs):
        """Constructs a variable using information from ``ast`` (such as label and line numbers).

        :param ast: AST node
        :type ast: :class:`jinja2.nodes.Node`
        """
        for k, v in kwargs.items():
            if v is None:
                del kwargs[k]
        kwargs = dict(cls._get_kwargs_from_ast(ast), **kwargs)
        return cls(**kwargs)

    @property
    def required(self):
        return not any([self.may_be_defined, self.used_with_default,
                        self.checked_as_defined, self.checked_as_undefined])

    def __eq__(self, other):
        return (
            type(self) is type(other) and
            self.linenos == other.linenos and
            self.label == other.label and
            self.constant == other.constant and
            self.used_with_default == other.used_with_default and
            self.checked_as_undefined == other.checked_as_undefined and
            self.checked_as_defined == other.checked_as_defined and
            self.required == other.required and
            self.value == other.value
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class Dictionary(Variable):
    """A dictionary.

    Implements some methods of Python :class:`dict`.

    .. automethod:: __setitem__
    .. automethod:: __getitem__
    .. automethod:: __delitem__
    .. automethod:: get
    .. automethod:: items
    .. automethod:: iteritems
    .. automethod:: keys
    .. automethod:: iterkeys
    .. automethod:: pop
    """

    def __init__(self, data=None, **kwargs):
        self.data = data or {}
        super(Dictionary, self).__init__(**kwargs)

    def __eq__(self, other):
        return super(Dictionary, self).__eq__(other) and self.data == other.data

    def __repr__(self):
        return pprint.pformat(self.data)

    def clone(self):
        rv = super(Dictionary, self).clone()
        rv.data = {}
        for k, v in _compat.iteritems(self.data):
            rv.data[k] = v.clone()
        return rv

    @classmethod
    def from_ast(cls, ast, data=None, **kwargs):
        kwargs = dict(cls._get_kwargs_from_ast(ast), **kwargs)
        return cls(data, **kwargs)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]

    def __contains__(self, key):
        return key in self.data

    def get(self, name, default=None):
        if name in self:
            return self[name]
        else:
            return default

    def items(self):
        return self.data.items()

    def iteritems(self):
        return _compat.iteritems(self.data)

    def keys(self):
        return self.data.keys()

    def iterkeys(self):
        return _compat.iterkeys(self.data)

    def pop(self, key, default=None):
        return self.data.pop(key, default)


class List(Variable):
    """A list which items are of the same type.

    .. attribute:: item

        A structure of list items, subclass of :class:`Variable`.
    """
    def __init__(self, item, **kwargs):
        self.item = item
        super(List, self).__init__(**kwargs)

    def __eq__(self, other):
        return super(List, self).__eq__(other) and self.item == other.item

    def __repr__(self):
        return pprint.pformat([self.item])

    def clone(self):
        rv = super(List, self).clone()
        rv.item = self.item.clone()
        return rv

    @classmethod
    def from_ast(cls, ast, item, **kwargs):
        kwargs = dict(cls._get_kwargs_from_ast(ast), **kwargs)
        return cls(item, **kwargs)


class Tuple(Variable):
    """A tuple.

    .. attribute:: items

        A :class:`tuple` of :class:`Variable` instances or ``None`` if the tuple items are unknown.

    .. attribute:: items

        Whether new elements can be added to the tuple in the process of merge or not.
    """
    def __init__(self, items, **kwargs):
        self.items = tuple(items) if items is not None else None
        self.may_be_extended = kwargs.pop('may_be_extended', False)
        super(Tuple, self).__init__(**kwargs)

    def __eq__(self, other):
        return super(Tuple, self).__eq__(other) and self.items == other.items

    def __repr__(self):
        return pprint.pformat(self.items)

    def clone(self):
        rv = super(Tuple, self).clone()
        rv.items = self.items and tuple(s.clone() for s in self.items)
        return rv

    @classmethod
    def from_ast(cls, ast, items, **kwargs):
        kwargs = dict(cls._get_kwargs_from_ast(ast), **kwargs)
        return cls(items, **kwargs)


class Scalar(Variable):
    """A scalar. Either string, number, boolean or ``None``."""
    def __repr__(self):
        return '<scalar>'


class String(Scalar):
    """A string."""
    def __repr__(self):
        return '<string>'


class Number(Scalar):
    """A number."""
    def __repr__(self):
        return '<number>'


class Boolean(Scalar):
    """A boolean."""
    def __repr__(self):
        return '<boolean>'


class Unknown(Variable):
    """A variable which type is unknown."""
    def __repr__(self):
        return '<unknown>'



# coding: utf-8
"""
jinja2schema.visitors.stmt
~~~~~~~~~~~~~~~~~~~~~~~~~~

Statement is an instance of :class:`jinja2.nodes.Stmt`.
Statement visitors return :class:`.models.Dictionary` of structures of variables used within the statement.
"""
import functools

from jinja2 import nodes, Environment, PackageLoader
from jinja2schema.config import default_config

from ..model import Scalar, Dictionary, List, Unknown, Tuple, Boolean
from ..macro import Macro
from ..mergers import merge, merge_many
from ..exceptions import InvalidExpression
from .._compat import iteritems, izip, zip_longest
from .expr import Context, visit_expr
from .util import visit_many


stmt_visitors = {}


def visits_stmt(node_cls):
    """Decorator that registers a function as a visitor for ``node_cls``.

    :param node_cls: subclass of :class:`jinja2.nodes.Stmt`
    """
    def decorator(func):
        stmt_visitors[node_cls] = func
        @functools.wraps(func)
        def wrapped_func(ast, macroses=None, config=default_config, child_blocks=None):
            assert isinstance(ast, node_cls)
            return func(ast, macroses, config, child_blocks)
        return wrapped_func
    return decorator


def visit_stmt(ast, macroses=None, config=default_config, child_blocks=None):
    """Returns a structure of ``ast``.

    :param ast: instance of :class:`jinja2.nodes.Stmt`
    :returns: :class:`.model.Dictionary`
    """
    visitor = stmt_visitors.get(type(ast))
    if not visitor:
        for node_cls, visitor_ in iteritems(stmt_visitors):
            if isinstance(ast, node_cls):
                visitor = visitor_
    if not visitor:
        raise Exception('stmt visitor for {0} is not found'.format(type(ast)))
    return visitor(ast, macroses, config)


@visits_stmt(nodes.For)
def visit_for(ast, macroses=None, config=default_config, child_blocks=None):
    body_struct = visit_many(ast.body, macroses, config, predicted_struct_cls=Scalar)
    else_struct = visit_many(ast.else_, macroses, config, predicted_struct_cls=Scalar)

    if 'loop' in body_struct:
        # exclude a special `loop` variable from the body structure
        del body_struct['loop']

    if isinstance(ast.target, nodes.Tuple):
        target_struct = Tuple.from_ast(
            ast.target,
            [body_struct.pop(item.name, Unknown.from_ast(ast.target))
             for item in ast.target.items])
    else:
        target_struct = body_struct.pop(ast.target.name, Unknown.from_ast(ast))

    iter_rtype, iter_struct = visit_expr(
        ast.iter,
        Context(
            return_struct_cls=Unknown,
            predicted_struct=List.from_ast(ast, target_struct)),
        macroses, config)

    merge(iter_rtype, List(target_struct))

    return merge_many(iter_struct, body_struct, else_struct)


@visits_stmt(nodes.If)
def visit_if(ast, macroses=None, config=default_config, child_blocks=None):
    if config.BOOLEAN_CONDITIONS:
        test_predicted_struct = Boolean.from_ast(ast.test)
    else:
        test_predicted_struct = Unknown.from_ast(ast.test)
    test_rtype, test_struct = visit_expr(
        ast.test, Context(predicted_struct=test_predicted_struct), macroses, config)
    if_struct = visit_many(ast.body, macroses, config, predicted_struct_cls=Scalar)
    else_struct = visit_many(ast.else_, macroses, config, predicted_struct_cls=Scalar) if ast.else_ else Dictionary()
    struct = merge_many(test_struct, if_struct, else_struct)

    for var_name, var_struct in iteritems(test_struct):
        if var_struct.checked_as_defined or var_struct.checked_as_undefined:
            if var_struct.checked_as_undefined:
                lookup_struct = if_struct
            elif var_struct.checked_as_defined:
                lookup_struct = else_struct
            struct[var_name].may_be_defined = (lookup_struct and
                                               var_name in lookup_struct and
                                               lookup_struct[var_name].constant)
            struct[var_name].checked_as_defined = test_struct[var_name].checked_as_defined and (
                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
            )
            struct[var_name].checked_as_undefined = test_struct[var_name].checked_as_undefined and (
                not lookup_struct or not var_name in lookup_struct or lookup_struct[var_name].constant
            )
    return struct


@visits_stmt(nodes.Assign)
def visit_assign(ast, macroses=None, config=default_config, child_blocks=None):
    struct = Dictionary()
    if (isinstance(ast.target, nodes.Name) or
            (isinstance(ast.target, nodes.Tuple) and isinstance(ast.node, nodes.Tuple))):
        variables = []
        if not (isinstance(ast.target, nodes.Tuple) and isinstance(ast.node, nodes.Tuple)):
            variables.append((ast.target.name, ast.node))
        else:
            if len(ast.target.items) != len(ast.node.items):
                raise InvalidExpression(ast, 'number of items in left side is different '
                                             'from right side')
            for name_ast, var_ast in izip(ast.target.items, ast.node.items):
                variables.append((name_ast.name, var_ast))
        for var_name, var_ast in variables:
            var_rtype, var_struct = visit_expr(var_ast, Context(predicted_struct=Unknown.from_ast(var_ast)), macroses, config)
            var_rtype.constant = True
            var_rtype.label = var_name
            struct = merge_many(struct, var_struct, Dictionary({
                var_name: var_rtype,
            }))
        return struct
    elif isinstance(ast.target, nodes.Tuple):
        tuple_items = []
        for name_ast in ast.target.items:
            var_struct = Unknown.from_ast(name_ast, constant=True)
            tuple_items.append(var_struct)
            struct = merge(struct, Dictionary({name_ast.name: var_struct}))
        var_rtype, var_struct = visit_expr(
            ast.node, Context(return_struct_cls=Unknown, predicted_struct=Tuple(tuple_items)), macroses, config)
        return merge(struct, var_struct)
    else:
        raise InvalidExpression(ast, 'unsupported assignment')


@visits_stmt(nodes.Output)
def visit_output(ast, macroses=None, config=default_config, child_blocks=None):
    return visit_many(ast.nodes, macroses, config, predicted_struct_cls=Scalar)


@visits_stmt(nodes.Macro)
def visit_macro(ast, macroses=None, config=default_config, child_blocks=None):
    # XXX the code needs to be refactored
    args = []
    kwargs = []
    body_struct = visit_many(ast.body, macroses, config, predicted_struct_cls=Scalar)

    for i, (arg, default_value_ast) in enumerate(reversed(list(zip_longest(reversed(ast.args),
                                                                           reversed(ast.defaults)))), start=1):
        has_default_value = bool(default_value_ast)
        if has_default_value:
            default_rtype, default_struct = visit_expr(
                default_value_ast, Context(predicted_struct=Unknown()), macroses, config)
        else:
            default_rtype = Unknown(linenos=[arg.lineno])
        default_rtype.constant = False
        default_rtype.label = 'argument "{0}"'.format(arg.name) if has_default_value else 'argument #{0}'.format(i)
        if arg.name in body_struct:
            default_rtype = merge(default_rtype, body_struct[arg.name])  # just to make sure
        default_rtype.linenos = [ast.lineno]
        if has_default_value:
            kwargs.append((arg.name, default_rtype))
        else:
            args.append((arg.name, default_rtype))
    macroses[ast.name] = Macro(ast.name, args, kwargs)

    tmp = dict(args)
    tmp.update(dict(kwargs))
    args_struct = Dictionary(tmp)
    for arg_name, arg_type in args:
        args_struct[arg_name] = arg_type

    for arg in args_struct.iterkeys():
        body_struct.pop(arg, None)
    return body_struct


@visits_stmt(nodes.Block)
def visit_block(ast, macroses=None, config=default_config):
    return visit_many(ast.body, macroses, config)


@visits_stmt(nodes.Include)
def visit_include(ast, macroses=None, config=default_config, child_blocks=None):
    template = get_inherited_template(config, ast)
    return visit_many(template.body, macroses, config)


@visits_stmt(nodes.Extends)
def visit_extends(ast, macroses=None, config=default_config, child_blocks=None):
    template = get_inherited_template(config, ast)
    if not child_blocks:
        return visit_many(template.body, macroses, config)
    return visit_many(get_correct_nodes(child_blocks, template.body), None, config)


def get_inherited_template(config, ast):
    env = Environment(loader=PackageLoader(config.PACKAGE_NAME, config.TEMPLATE_DIR))
    return env.parse(env.loader.get_source(env, ast.template.value)[0])


def separate_template_blocks(template, blocks, template_nodes):
    for node in template:
        if isinstance(node, nodes.Block):
            blocks.append(node)
        else:
            template_nodes.append(node)
    return blocks, template_nodes


def get_correct_nodes(child_blocks, template):
    parent_blocks, nodes = separate_template_blocks(template, [], [])
    child_block_names = [c.name for c in child_blocks]
    blocks = child_blocks + parent_blocks
    for parent_block in parent_blocks:
        if parent_block.name in child_block_names:
            blocks.remove(parent_block)
    return blocks + nodes



# coding: utf-8
"""
jinja2schema.config
~~~~~~~~~~~~~~~~~~~
"""


class Config(object):
    """Configuration."""

    TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE = 'dictionary'
    """Possible values: ``"dictionary"`` or ``"list""``.

    For example, in the expression ``xs[a]`` variable ``xs`` may be a list as well as a dictionary.
    This setting is used to choose between a dictionary and a list when the variable is
    being indexed with another variable.
    """

    TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE = 'list'
    """Possible values: ``"dictionary"``, ``"list"`` or ``"tuple"``.

    For example, in the expression ``xs[2]`` variable ``xs`` may be a list as well as a dictionary or a tuple.
    This setting is used to choose between a dictionary, a tuple and a list when the variable is
    being indexed with an integer.
    """

    BOOLEAN_CONDITIONS = False
    """Whether or not to consider conditions in ``if`` statements as boolean.

    If this variable is not set, ``xs`` variable in template ``{% if xs %}{% endif %}`` will have
    unknown structure. If this variable is set, ``xs`` will be a boolean.
    """

    PACKAGE_NAME = ''
    """Name of the package where you want to load templates from.

    This configuration is for if you are using includes in your jinja templates. This tells jinja
    where to look to be able to load the included template from. If you do not plan on using ``includes``
    this configuration is not needed.
    """

    TEMPLATE_DIR = 'templates'
    """Name of the directory where you want to load templates from. Defaulted to ``templates``

    This configuration is for if you are using includes in your jinja templates. This tells jinja
    which directoy to look to be able to load the included template from. If you do not plan on using ``includes``
    this configuration is not needed.
    """

    def __init__(self,
                 TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE='dictionary',
                 TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE='list',
                 BOOLEAN_CONDITIONS=False,
                 PACKAGE_NAME='',
                 TEMPLATE_DIR='templates'):
        if TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE not in ('dictionary', 'list'):
            raise ValueError('TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE must be'
                             'either "dictionary" or "list"')
        if TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE not in ('dictionary', 'list', 'tuple'):
            raise ValueError('TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE must be'
                             'either "dictionary", "tuple" or "list"')
        self.TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE = TYPE_OF_VARIABLE_INDEXED_WITH_INTEGER_TYPE
        self.TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE = TYPE_OF_VARIABLE_INDEXED_WITH_VARIABLE_TYPE
        self.BOOLEAN_CONDITIONS = BOOLEAN_CONDITIONS
        self.PACKAGE_NAME = PACKAGE_NAME
        self.TEMPLATE_DIR = TEMPLATE_DIR


default_config = Config()



# coding: utf-8
"""
Some py2/py3 compatibility support based on a stripped down
version of six so we don't have to depend on a specific version
of it.
"""
import sys

PY2 = sys.version_info[0] == 2
PYPY = hasattr(sys, 'pypy_translation_info')
_identity = lambda x: x


if not PY2:
    unichr = chr
    range_type = range
    text_type = str
    string_types = (str,)

    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())

    import pickle
    from io import BytesIO, StringIO
    NativeStringIO = StringIO

    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    ifilter = filter
    imap = map
    izip = zip
    intern = sys.intern

    get_next = lambda x: x.__next__

    from itertools import zip_longest
else:
    unichr = unichr
    text_type = unicode
    range_type = xrange
    string_types = (str, unicode)

    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()

    import cPickle as pickle
    from cStringIO import StringIO as BytesIO, StringIO
    NativeStringIO = BytesIO

    exec('def reraise(tp, value, tb=None):\n raise tp, value, tb')

    from itertools import imap, izip, ifilter
    intern = intern

    get_next = lambda x: x.next

    from itertools import izip_longest as zip_longest



# coding: utf-8
"""
jinja2schema.exceptions
~~~~~~~~~~~~~~~~~~~~~~~
"""
from jinja2schema.model import Scalar


class InferException(Exception):
    """Base class for jinja2schema exceptions."""


class MergeException(InferException):
    """Conflict of merging two structures.

    .. attribute:: fst

        :class:`Variable`

    .. attribute:: snd

        :class:`Variable`
    """
    def __init__(self, fst, snd):
        self.fst = fst
        self.snd = snd

    def __str__(self):
        get_label = lambda s: 'unnamed variable' if s.label is None else 'variable "{0}"'.format(s.label)
        def get_usage(s):
            return s.__class__.__name__.lower()
        get_linenos = lambda s: ', '.join(map(str, s.linenos))
        return ('{fst_label} (used as {fst_usage} on lines {fst_linenos}) conflicts with '
                '{snd_label} (used as {snd_usage} on lines: {snd_linenos})').format(
                    fst_label=get_label(self.fst), snd_label=get_label(self.snd),
                    fst_usage=get_usage(self.fst), snd_usage=get_usage(self.snd),
                    fst_linenos=get_linenos(self.fst), snd_linenos=get_linenos(self.snd))


class UnexpectedExpression(InferException):
    """Raised when a visitor was expecting compatibility with :attr:`expected_struct`,
    but got :attr:`actual_ast` of structure :attr:`actual_struct`.

    Compatibility is checked by merging expected structure with actual one.

    .. attribute:: expected_struct

        expected :class:`.model.Variable`

    .. attribute:: actual_ast

        actual :class:`jinja2.nodes.Node`

    .. attribute:: actual_ast

        :class:`.model.Variable` described by ``actual_ast``
    """
    def __init__(self, expected_struct, actual_ast, actual_struct):
        self.expected_struct = expected_struct
        self.actual_ast = actual_ast
        self.actual_struct = actual_struct

    def __str__(self):
        return ('conflict on the line {lineno}\n'
                'got: AST node jinja2.nodes.{node} of structure {actual_struct}\n'
                'expected structure: {expected_struct}').format(
                    lineno=self.actual_ast.lineno,
                    node=self.actual_ast.__class__.__name__,
                    actual_struct=self.actual_struct,
                    expected_struct=self.expected_struct)


class InvalidExpression(InferException):
    """Raised when a template uses Jinja2 features that are not supported by the library
    or when a template contains incorrect expressions (i.e., such as applying ``divisibleby`` filter
    without an argument).

    .. attribute:: ast

        :class:`jinja2.nodes.Node` caused the exception
    """
    def __init__(self, ast, message):
        self.ast = ast
        self.message = message

    def __str__(self):
        return 'line {0}: {1}'.format(self.ast.lineno, self.message)